﻿private import std.socket, std.thread, std.stdio;
pragma(lib,"ws2_32.lib");

alias char[] string;

class ListenThread : Thread
{
	Socket listen_socket;
	
	this() {
		//
	}

	~this() {
		delete listen_socket;
	}

	int run() {
		listen_socket = new TcpSocket;
		listen_socket.bind( new InternetAddress(InternetAddress.ADDR_ANY, 1234) );
		listen_socket.listen(5);
		while (1) {
			Socket news = listen_socket.accept();
			if (!news)
				break;
			auto ht = new FileServer(news);
			ht.start();
		}
		return 1;
	}

	void stop() {
		listen_socket.close();
	}
}

class ClientThread : Thread
{
	Socket client_socket;

	this( Socket listen_socket ) {
		client_socket = listen_socket;
		assert(client_socket);
	}

	~this() {
		assert(client_socket);
		delete client_socket;
	}

	int run() {
		assert(client_socket);
		string headers;
		while (1) {
			char buf[1024];
			int error = client_socket.receive( buf );
			if (error <= 0)
				break;
			headers ~= buf[0..error];
			if ((headers.length >= 2 && headers[$-2..$]=="\n\n") ||
				(headers.length >= 4 && headers[$-4..$]=="\r\n\r\n")) {
				// got headers
				serve(headers);
				break;
			}
		}	
		client_socket.close();
		return 1;
	}

	void serve( string headers ) {
const string response = "HTTP/1.0 200 OK\r\nContent-Type: text/plain\r\n\r\n";
		client_socket.send( response ~ headers );
	}
}


import std.string, std.file, std.uri;

class FileServer : ClientThread
{
	string			verb, uri, ver;
	string[string]	fields;

	this( Socket listen_socket ) {
		super(listen_socket);
	}

	void serve( string headers ) {
		string[] rows = headers.split("\n");
		if (rows.length == 0)
			throw new Exception("empty headers");
		// split the first line by spaces
		string[] http = rows[0].split();
		if (http.length != 3)
			throw new Exception("misformed http header");
		verb = http[0];
		uri = std.uri.decodeComponent( http[1] );
		ver = http[2];
		// split the rest by colon
		foreach(row; rows[1..$]) {
			int colon = row.find(':');
			if (colon >= 0) {
				string key = row[0..colon];
				string value = strip( row[colon+1..$] );
				fields[ key ] = value;
			}
		}
		// do something
		switch (verb) {
		case "GET":
			try {
				if (isdir(uri))
					uri ~= '/';
				if (uri[$-1]=='/')   {
const string response = "HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n";
					client_socket.send(response);
//					uri ~= "index.html";	//default document
					char[][] files = listdir(uri);
					foreach(file;files)
						client_socket.send("<a href='"~std.uri.encode(uri~file)~"'>"~file~"</a><br>");
				} else {
					char[] contents = cast(char[])read(uri);
const string response = "HTTP/1.0 200 OK\r\nContent-Type: text/plain\r\n\r\n";
					client_socket.send( response ~ contents );
				}
			}
			catch(Object o) {
const string response = "HTTP/1.0 404 Not found\r\n\r\n";
				client_socket.send( response ~ "<html><body>"~o.toString~"</body></html>" );
			}
			break;
		default:
			throw new Exception("unsupported verb");
		}
	}
}

int main()
{
	try {
		auto t = new ListenThread;
		t.start();
		writefln("Waiting for incoming connections...");
		getchar();
		writefln("Closing connections...");
		t.stop();
	} catch( Object o ) {
		writefln( "%s", o.toString );
		getchar();
	}
    return 0;
}
